from django.db import models
from django.utils.translation import ugettext_lazy as _
from django.utils import timezone

from cms.models.pluginmodel import CMSPlugin

from djangocms_text_ckeditor.fields import HTMLField


class NewsItem(models.Model):
    '''
    Each NewsItem will be posted on the news page, with the most recent news first.
    Alerts can be posted with special CSS to make them stand out.
    The most recent news item may be shown on the home page.
    '''

    title = models.CharField(_('Title'),max_length=200,help_text=_('Title the news item. Make sure this is descriptive.'))

    content = HTMLField(_('Content'),help_text=_('Insert news content here.'))

    alert = models.BooleanField(_('Alert'),default=False, help_text=_('Alerts (such as cancellations) may be displayed in an emphasized fashion.'))

    pinThis = models.BooleanField(_('Pinned'), default=False, help_text=_('If this field is set, then the news item will continue to show up on the main page until someone unpins it.'))

    draft = models.BooleanField(_('Draft status'),default=False, help_text=_('Check to hide from publication'))
    hideThis = models.BooleanField(_('Hidden'), default=False, help_text=_('If for some reason you need to make a news item invisible, this will do that.'))

    creationDate = models.DateTimeField(_('Creation date'),auto_now_add=True)
    modifiedDate = models.DateTimeField(_('Last modified date'),auto_now=True,help_text=_('Time of most recent edit'))
    publicationDate = models.DateTimeField(_('Publication date'),default=timezone.now)

    def __str__(self):
        return self.title

    class Meta:
        verbose_name = _('News item')
        verbose_name_plural = _('News items')
        ordering = ('-pinThis','-publicationDate',)


class LatestNewsPluginModel(CMSPlugin):
    maxItems = models.PositiveSmallIntegerField(_('Maximum number of items to show'),default=1)
    daysBefore = models.PositiveSmallIntegerField(_('Published since (days ago)'),help_text=_('Leave blank for no limit.'),null=True,blank=True)
    alertOnly = models.BooleanField(_('Show Alerts Only'),default=False)
    ignorePins = models.BooleanField(_('Ignore Pinned Item Precedence'),help_text=_('By default, pinned items are shown first, regardless of publication date. Checking this box overrides that behavior.'),default=False)
    template = models.CharField(_('Template'),max_length=250,null=True,blank=True)

    def get_short_description(self):
        desc = self.id
        choices = getattr(self.get_plugin_class(),'template_choices',[])
        choice_name = [x[1] for x in choices if x[0] == self.template]
        if choice_name:
            desc = choice_name[0]
        elif self.template:
            desc = self.template
        return desc



# -*- coding: utf-8 -*-
# Generated by Django 1.10.7 on 2017-06-19 22:55
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0005_auto_20170614_1607'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='dancetypelevel',
            name='requiresAudition',
        ),
    ]



from django.contrib.sitemaps import Sitemap

from .models import Event


class EventSitemap(Sitemap):
    changefreq = "monthly"
    protocol = "https"

    def items(self):
        return Event.objects.exclude(status=Event.RegStatus.hidden)

    def lastmod(self, obj):
        return obj.modified



from django.contrib import admin
from django.forms import ModelForm, SplitDateTimeField, HiddenInput, RadioSelect, ModelMultipleChoiceField
from django.utils.safestring import mark_safe
from django.core.urlresolvers import reverse
from django.utils.translation import ugettext_lazy as _
from django.template.response import SimpleTemplateResponse
from django.contrib.contenttypes.models import ContentType
from django.http import HttpResponseRedirect

from calendar import month_name
from polymorphic.admin import PolymorphicParentModelAdmin, PolymorphicChildModelAdmin, PolymorphicChildModelFilter
from cms.admin.placeholderadmin import FrontendEditableAdminMixin
import json
import six
from dal import autocomplete

from .models import EventSession, Event, PublicEventCategory, Series, SeriesCategory, PublicEvent, EventOccurrence, SeriesTeacher, StaffMember, Instructor, SubstituteTeacher, Registration, TemporaryRegistration, EventRegistration, TemporaryEventRegistration, ClassDescription, CustomerGroup, Customer, Location, PricingTier, DanceRole, DanceType, DanceTypeLevel, EmailTemplate, EventStaffMember, EventStaffCategory, EventRole, Invoice, InvoiceItem, Room
from .constants import getConstant
from .forms import LocationWithDataWidget


######################################
# Admin action for repeating events


def repeat_events(modeladmin, request, queryset):
    selected = request.POST.getlist(admin.ACTION_CHECKBOX_NAME)
    ct = ContentType.objects.get_for_model(queryset.model)
    return HttpResponseRedirect(reverse('repeatEvents') + "?ct=%s&ids=%s" % (ct.pk, ",".join(selected)))


repeat_events.short_description = _('Duplicate selected events')


######################################
# Inline classes


class EventRoleInline(admin.TabularInline):
    model = EventRole
    extra = 1
    classes = ['collapse']

    verbose_name = _('Event-specific dance role')
    verbose_name_plural = _('Event-specific dance roles (override default)')


class SeriesTeacherInlineForm(ModelForm):

    def __init__(self, *args, **kwargs):
        super(SeriesTeacherInlineForm,self).__init__(*args,**kwargs)

        self.fields['staffMember'].label = _('Instructor')
        self.fields['category'].initial = getConstant('general__eventStaffCategoryInstructor').id

        # Impose restrictions on new records, but not on existing ones.
        if not kwargs.get('instance',None):
            # Filter out retired teachers
            self.fields['staffMember'].queryset = Instructor.objects.exclude(status__in=[Instructor.InstructorStatus.retired,Instructor.InstructorStatus.hidden,Instructor.InstructorStatus.retiredGuest])
        else:
            self.fields['staffMember'].queryset = Instructor.objects.all()

        self.fields['staffMember'].queryset = self.fields['staffMember'].queryset.order_by('status','firstName','lastName')

    class Meta:
        widgets = {
            'category': HiddenInput(),
        }


class SeriesTeacherInline(admin.StackedInline):
    model = SeriesTeacher
    form = SeriesTeacherInlineForm
    exclude = ('replacedStaffMember','occurrences','submissionUser')
    extra = 1

    def save_model(self,request,obj,form,change):
        obj.replacedStaffMember = None
        obj.occurrences = obj.event.eventoccurrence_set.all()
        obj.submissionUser = request.user
        obj.save()


class SubstituteTeacherInlineForm(ModelForm):

    def __init__(self, *args, **kwargs):
        super(SubstituteTeacherInlineForm,self).__init__(*args,**kwargs)

        self.fields['staffMember'].label = _('Instructor')
        self.fields['replacedStaffMember'].required = True
        self.fields['category'].initial = getConstant('general__eventStaffCategorySubstitute').id

        # Impose restrictions on new records, but not on existing ones.
        if not kwargs.get('instance',None):
            # Filter out retired teachers
            self.fields['staffMember'].queryset = Instructor.objects.exclude(status__in=[Instructor.InstructorStatus.retired,Instructor.InstructorStatus.hidden])
        else:
            self.fields['staffMember'].queryset = Instructor.objects.all()

        self.fields['staffMember'].queryset = self.fields['staffMember'].queryset.order_by('status','firstName','lastName')

    class Meta:
        widgets = {
            'category': HiddenInput(),
        }


class SubstituteTeacherInline(admin.StackedInline):
    model = SubstituteTeacher
    form = SubstituteTeacherInlineForm
    exclude = ('submissionUser',)
    extra = 0

    def formfield_for_foreignkey(self, db_field, request=None, **kwargs):
        field = super(SubstituteTeacherInline, self).formfield_for_foreignkey(db_field, request, **kwargs)

        if db_field.name == 'replacedStaffMember':
            if request._obj_ is not None:
                # set the query set to whatever you like
                field.queryset = SeriesTeacher.objects.filter(event=request._obj_)
            else:
                field.queryset = field.queryset.none()
        return field

    def formfield_for_manytomany(self, db_field, request=None, **kwargs):
        field = super(SubstituteTeacherInline, self).formfield_for_foreignkey(db_field, request, **kwargs)

        if db_field.name == 'occurrences':
            if request._obj_ is not None:
                # set the query set to whatever you like
                field.queryset = EventOccurrence.objects.filter(event=request._obj_)
            else:
                field.queryset = field.queryset.none()
        return field

    def save_model(self,request,obj,form,change):
        obj.submissionUser = request.user
        obj.save()


class EventStaffMemberInline(admin.StackedInline):
    model = EventStaffMember
    exclude = ('submissionUser','replacedStaffMember')
    extra = 0

    def formfield_for_manytomany(self, db_field, request=None, **kwargs):
        field = super(EventStaffMemberInline, self).formfield_for_foreignkey(db_field, request, **kwargs)

        if db_field.name == 'occurrences':
            if request._obj_ is not None:
                # set the query set to whatever you like
                field.queryset = EventOccurrence.objects.filter(event=request._obj_)
            else:
                field.queryset = field.queryset.none()
        return field

    def save_model(self,request,obj,form,change):
        obj.submissionUser = request.user
        obj.save()


class EventRegistrationInline(admin.StackedInline):
    model = EventRegistration
    extra = 0
    fields = ['event','role',('checkedIn','cancelled'),'dropIn','price','netPrice']
    readonly_fields = ['event','price','netPrice','dropIn']

    # These ensure that registration changes happen through the regular registration
    # process.
    def has_add_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False


class EventOccurrenceInlineForm(ModelForm):
    WIDGET_FORMATS = ['%I:%M%p','%I:%M %p','%I:%M','%H:%M:%S','%H:%M']

    startTime = SplitDateTimeField(required=True,label=_('Start Date/Time'),input_time_formats=WIDGET_FORMATS)
    endTime = SplitDateTimeField(required=True,label=_('End Date/Time'),input_time_formats=WIDGET_FORMATS)


class EventOccurrenceInline(admin.TabularInline):
    model = EventOccurrence
    form = EventOccurrenceInlineForm
    extra = 1

    class Media:
        js = ('timepicker/jquery.timepicker.min.js','jquery-ui/jquery-ui.min.js','datepair/datepair.min.js','moment/moment.min.js','datepair/jquery.datepair.min.js','js/eventadmin_pickers.js')
        css = {'all':('timepicker/jquery.timepicker.css','jquery-ui/jquery-ui.min.css',)}


######################################
# Registration related admin classes


class InvoiceItemInline(admin.StackedInline):
    model = InvoiceItem
    extra = 0
    add_fields = [('description','grossTotal','total','taxes','fees','adjustments'),]
    fields = ['id',('description','grossTotal','total','taxes','fees','adjustments'),]
    add_readonly_fields = ['fees',]
    readonly_fields = ['id','grossTotal','total','taxes','fees']

    # This ensures that InvoiceItems are not deleted except through
    # the regular registration process.  Invoice items can still be
    # manually added.
    def has_delete_permission(self, request, obj=None):
        return False

    def get_readonly_fields(self, request, obj=None):
        if not obj:
            return self.add_readonly_fields
        return self.readonly_fields

    def get_fields(self, request, obj=None):
        if not obj:
            return self.add_fields
        return super(InvoiceItemInline, self).get_fields(request, obj)


@admin.register(Invoice)
class InvoiceAdmin(admin.ModelAdmin):
    inlines = [InvoiceItemInline,]
    list_display = ['id','recipientInfo','status', 'total','netRevenue','outstandingBalance','creationDate','modifiedDate','links']
    list_filter = ['status', 'paidOnline', 'creationDate', 'modifiedDate']
    search_fields = ['id','comments']
    ordering = ['-modifiedDate',]
    readonly_fields = ['id','recipientInfo','total','adjustments','taxes','fees','netRevenue','outstandingBalance','creationDate','modifiedDate','links','submissionUser','collectedByUser']
    view_on_site = True

    def emailNotification(self, request, queryset):
        # Allows use of the email view to contact specific customers.
        selected = request.POST.getlist(admin.ACTION_CHECKBOX_NAME)
        return HttpResponseRedirect(reverse('sendInvoiceNotifications') + "?invoices=%s" % (",".join(selected)))
    emailNotification.short_description = _('Send email notifications for selected invoices')

    actions = ['emailNotification',]

    def recipientInfo(self,obj):
        if obj.firstName and obj.lastName and obj.email:
            return '%s %s: %s' % (obj.firstName, obj.lastName, obj.email)
        elif obj.email:
            return obj.email
    recipientInfo.short_description = _('Recipient')

    def viewInvoiceLink(self,obj):
        if obj.id:
            change_url = reverse('viewInvoice', args=(obj.id,))
            return mark_safe('<a class="btn btn-outline-secondary" href="%s">View Invoice</a>' % (change_url,))
    viewInvoiceLink.allow_tags = True
    viewInvoiceLink.short_description = _('Invoice')

    def notificationLink(self,obj):
        if obj.id:
            change_url = reverse('sendInvoiceNotifications', args=(obj.id,))
            return mark_safe('<a class="btn btn-outline-secondary" href="%s">Notify Recipient</a>' % (change_url,))
    notificationLink.allow_tags = True
    notificationLink.short_description = _('Invoice notification')

    def finalRegistrationLink(self,obj):
        if obj.finalRegistration:
            change_url = reverse('admin:core_registration_change', args=(obj.finalRegistration.id,))
            return mark_safe('<a class="btn btn-outline-secondary" href="%s">Registration</a>' % (change_url,))
    finalRegistrationLink.allow_tags = True
    finalRegistrationLink.short_description = _('Final registration')

    def temporaryRegistrationLink(self,obj):
        if obj.temporaryRegistration:
            change_url = reverse('admin:core_temporaryregistration_change', args=(obj.temporaryRegistration.id,))
            return mark_safe('<a class="btn btn-outline-secondary" href="%s">Temporary Registration</a>' % (change_url,))
    temporaryRegistrationLink.allow_tags = True
    temporaryRegistrationLink.short_description = _('Temporary registration')

    def links(self,obj):
        return mark_safe(''.join([
            self.viewInvoiceLink(obj) or '',
            self.notificationLink(obj) or '',
            self.temporaryRegistrationLink(obj) or '',
            self.finalRegistrationLink(obj) or '',
        ]))
    links.allow_tags = True
    links.short_description = _('Links')

    def save_model(self, request, obj, form, change):
        if not change:
            obj.submissionUser = request.user
        super(InvoiceAdmin, self).save_model(request, obj, form, change)

    def save_related(self, request, form, formsets, change):
        '''
        Update the Invoice object after saving so that the invoice lines
        match the invoice item lines.
        '''
        super(InvoiceAdmin, self).save_related(request, form, formsets, change)
        invoice = form.instance
        for attr in ['grossTotal','total','adjustments','taxes','fees']:
            setattr(invoice,attr,sum([getattr(x,attr) for x in invoice.invoiceitem_set.all()]))
        invoice.save()

    def get_fieldsets(self,request,obj=None):
        if not obj:
            return self.add_fieldsets
        else:
            return super(InvoiceAdmin,self).get_fieldsets(request,obj)

    fieldsets = (
        (None, {
            'fields': ('id',('firstName','lastName','email'),'comments','status','amountPaid','outstandingBalance','links'),
        }),
        (_('Financial Details'), {
            'classes': ('collapse',),
            'fields': ('total','adjustments','taxes','fees','netRevenue'),
        }),
        (_('Dates'), {
            'classes': ('collapse',),
            'fields': ('creationDate','modifiedDate'),
        }),
        (_('Additional data'), {
            'classes': ('collapse',),
            'fields': ('submissionUser','collectedByUser','data'),
        }),
    )

    add_fieldsets = (
        (None, {
            'fields': (('firstName','lastName','email'),'comments','status','amountPaid',),
        }),
        (_('Additional data'), {
            'classes': ('collapse',),
            'fields': ('data',),
        }),
    )


@admin.register(Registration)
class RegistrationAdmin(admin.ModelAdmin):
    inlines = [EventRegistrationInline]
    list_display = ['customer','dateTime','priceWithDiscount','student']
    list_filter = ['dateTime','student','invoice__paidOnline']
    search_fields = ['=customer__first_name','=customer__last_name','customer__email']
    ordering = ('-dateTime',)
    fields = ('customer_link','priceWithDiscount','student','dateTime','comments','howHeardAboutUs')
    readonly_fields = ('customer_link',)

    def customer_link(self,obj):
        change_url = reverse('admin:core_customer_change', args=(obj.customer.id,))
        return mark_safe('<a href="%s">%s</a>' % (change_url, obj.customer))

    customer_link.allow_tags = True
    customer_link.short_description = _("Customer")

    def save_formset(self, request, form, formset, change):
        instances = formset.save(commit=False)
        for obj in formset.deleted_objects:
            obj.delete()
        for instance in instances:
            if not hasattr(instance,'customer'):
                instance.customer = instance.registration.customer
            instance.save()
        formset.save_m2m()


class TemporaryEventRegistrationInline(admin.StackedInline):
    model = TemporaryEventRegistration
    extra = 0

    # These ensure that registration changes happen through the regular registration
    # process.
    def has_add_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False


@admin.register(TemporaryRegistration)
class TemporaryRegistrationAdmin(admin.ModelAdmin):
    inlines = [TemporaryEventRegistrationInline,]

    list_display = ('__str__','student','dateTime','expirationDate')
    search_fields = ('=firstName','=lastName','email')
    list_filter = ('dateTime','expirationDate',)


######################################
# Miscellaneous Admin classes

@admin.register(ClassDescription)
class ClassDescriptionAdmin(FrontendEditableAdminMixin, admin.ModelAdmin):
    list_display = ['title','danceTypeLevel','oneTimeSeries']
    list_filter = ('danceTypeLevel','oneTimeSeries')
    list_editable = ('oneTimeSeries',)
    search_fields = ('title',)
    prepopulated_fields = {"slug": ("title",)}


class CustomerRegistrationInline(admin.StackedInline):
    model = Registration
    fields = ('registration_link','eventregistration_list')
    readonly_fields = ('registration_link','eventregistration_list')
    extra = 0

    def registration_link(self,obj):
        change_url = reverse('admin:core_registration_change', args=(obj.id,))
        if obj.dateTime:
            return mark_safe('%s: <a href="%s">%s</a>' % (obj.dateTime.strftime('%b. %d, %Y'), change_url, obj.__str__()))
        else:
            return mark_safe('<a href="%s">%s</a>' % (change_url, obj.__str__()))

    registration_link.short_description = _('Registration')
    registration_link.allow_tags = True

    def eventregistration_list(self,obj):
        eregs = obj.eventregistration_set.all()
        if not eregs:
            return ''
        return_string = '<ul>'

        for ereg in eregs:
            this_string = '<li>'
            if ereg.cancelled:
                this_string += '<em>%s</em> ' % _('CANCELLED:')
            if ereg.dropIn:
                this_string += '<em>%s</em> ' % _('DROP-IN:')
            if ereg.event.month:
                this_string += '%s %s, %s</li>' % (month_name[ereg.event.month], ereg.event.year, ereg.event.name)
            else:
                this_string += '%s</li>' % ereg.event.name
            return_string += this_string
        return return_string
    eventregistration_list.short_description = _('Event Registrations')
    eventregistration_list.allow_tags = True

    # Prevents adding new registrations without going through
    # the standard registration process
    def has_add_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False


@admin.register(Customer)
class CustomerAdmin(admin.ModelAdmin):
    list_display = ('fullName','numClassSeries','numPublicEvents')
    search_fields = ('=first_name','=last_name','email')
    readonly_fields = ('data','numClassSeries','numPublicEvents')

    fieldsets = (
        (None, {
            'fields': (('first_name','last_name'),'email','phone','user',)
        }),
        (_('Groups'), {
            'fields': ('groups',)
        }),
        (_('Additional Customer Data'), {
            'classes': ('collapse',),
            'fields': (('numClassSeries','numPublicEvents',),'data',),
        }),
    )

    def emailCustomers(self, request, queryset):
        # Allows use of the email view to contact specific customers.
        selected = request.POST.getlist(admin.ACTION_CHECKBOX_NAME)
        return HttpResponseRedirect(reverse('emailStudents') + "?customers=%s" % (",".join(selected)))
    emailCustomers.short_description = _('Email selected customers')

    inlines = [CustomerRegistrationInline,]
    actions = ['emailCustomers']


class CustomerGroupAdminForm(ModelForm):
    customers = ModelMultipleChoiceField(
        queryset=Customer.objects.all(),
        required=False,
        widget=autocomplete.ModelSelect2Multiple(
            url='autocompleteCustomer',
            attrs={
                # This will set the input placeholder attribute:
                'data-placeholder': _('Enter a customer name'),
                # This will set the yourlabs.Autocomplete.minimumCharacters
                # options, the naming conversion is handled by jQuery
                'data-minimum-input-length': 2,
                'data-max-results': 4,
                'class': 'modern-style',
            }
        )
    )

    def __init__(self, *args, **kwargs):
        super(CustomerGroupAdminForm, self).__init__(*args, **kwargs)
        if self.instance.pk:
            self.initial['customers'] = self.instance.customer_set.values_list('pk', flat=True)

    def save(self, *args, **kwargs):
        instance = super(CustomerGroupAdminForm, self).save(*args, **kwargs)
        if instance.pk:
            instance.customer_set.clear()
            instance.customer_set.add(*self.cleaned_data['customers'])
        return instance

    class Meta:
        model = CustomerGroup
        exclude = []


@admin.register(CustomerGroup)
class CustomerGroupAdmin(admin.ModelAdmin):
    list_display = ('name','memberCount')
    readonly_fields = ('memberCount',)
    form = CustomerGroupAdminForm

    def emailCustomers(self, request, queryset):
        # Allows use of the email view to contact specific customer groups.
        selected = request.POST.getlist(admin.ACTION_CHECKBOX_NAME)
        return HttpResponseRedirect(reverse('emailStudents') + "?customergroup=%s" % (",".join(selected)))
    emailCustomers.short_description = _('Email selected customer groups')

    actions = ['emailCustomers']


@admin.register(EventSession)
class EventSessionAdmin(admin.ModelAdmin):
    list_display = ('name', 'startTime', 'endTime')
    ordering = ('startTime','name')
    readonly_fields = ('startTime','endTime')
    list_filter = ('startTime','endTime')
    prepopulated_fields = {'slug': ('name',)}

    fields = ('name', 'description', 'slug',('startTime','endTime'))


class RoomInline(admin.StackedInline):
    model = Room
    extra = 0
    fields = (('name', 'defaultCapacity'),'description')


@admin.register(Room)
class RoomAdmin(admin.ModelAdmin):
    inlines = []

    list_display = ('name','location','defaultCapacity')
    list_display_links = ('name',)
    list_editable = ('defaultCapacity',)
    list_filter = ('location',)

    ordering = ('location__name','name')

    fields = ('location','name','defaultCapacity','description')


@admin.register(Location)
class LocationAdmin(admin.ModelAdmin):
    inlines = [RoomInline,]

    list_display = ('name','address','city','orderNum','status')
    list_display_links = ('name',)
    list_editable = ('orderNum','status')
    list_filter = ('status','city')

    ordering = ('status','orderNum')

    def response_add(self, request, obj, post_url_continue=None):
        '''
        This just modifies the normal ModelAdmin process in order to
        pass capacity and room options for the added Location along with
        the location's name and ID.
        '''

        IS_POPUP_VAR = '_popup'
        TO_FIELD_VAR = '_to_field'

        if IS_POPUP_VAR in request.POST:
            to_field = request.POST.get(TO_FIELD_VAR)
            if to_field:
                attr = str(to_field)
            else:
                attr = obj._meta.pk.attname
            value = obj.serializable_value(attr)
            popup_response_data = json.dumps({
                'value': six.text_type(value),
                'obj': six.text_type(obj),
                # Add this extra data
                'defaultCapacity': obj.defaultCapacity,
                'roomOptions': json.dumps([{'id': x.id, 'name': x.name, 'defaultCapacity': x.defaultCapacity} for x in obj.room_set.all()]),
            })

            # Return a modified template
            return SimpleTemplateResponse('core/admin/location_popup_response.html', {
                'popup_response_data': popup_response_data,
            })

        # Otherwise just use the standard ModelAdmin method
        return super(LocationAdmin,self).response_add(request, obj, post_url_continue)

    def response_change(self, request, obj):
        '''
        This just modifies the normal ModelAdmin process in order to
        pass capacity and room options for the modified Location along with
        the location's name and ID.
        '''

        IS_POPUP_VAR = '_popup'
        TO_FIELD_VAR = '_to_field'

        if IS_POPUP_VAR in request.POST:
            to_field = request.POST.get(TO_FIELD_VAR)
            attr = str(to_field) if to_field else obj._meta.pk.attname
            # Retrieve the `object_id` from the resolved pattern arguments.
            value = request.resolver_match.args[0]
            new_value = obj.serializable_value(attr)
            popup_response_data = json.dumps({
                'action': 'change',
                'value': six.text_type(value),
                'obj': six.text_type(obj),
                'new_value': six.text_type(new_value),
                # Add this extra data
                'defaultCapacity': obj.defaultCapacity,
                'roomOptions': json.dumps([{'id': x.id, 'name': x.name, 'defaultCapacity': x.defaultCapacity} for x in obj.room_set.all()]),
            })

            # Return a modified template
            return SimpleTemplateResponse('core/admin/location_popup_response.html', {
                'popup_response_data': popup_response_data,
            })
        return super(LocationAdmin,self).response_change(request, obj)


@admin.register(PricingTier)
class PricingTierAdmin(admin.ModelAdmin):
    list_display = ('name','expired')
    list_filter = ('expired',)

    # Need to specify an empty list of inlines so that discounts app can add to the list if it is enabled.
    inlines = []


class EmailTemplateAdminForm(ModelForm):

    class Meta:
        model = EmailTemplate
        exclude = []

        widgets = {
            'richTextChoice': RadioSelect,
        }

    class Media:
        js = ('js/emailtemplate_contenttype.js',)


@admin.register(EmailTemplate)
class EmailTemplateAdmin(admin.ModelAdmin):
    form = EmailTemplateAdminForm

    list_display = ('name','richTextChoice','hideFromForm')
    list_filter = ('richTextChoice','groupRequired','hideFromForm')
    ordering = ('name',)

    fieldsets = (
        (None,{
            'fields': ('name','richTextChoice','subject',),
        }),
        (_('Plain text content'),{
            'fields': ('content',),
        }),
        (_('Rich text HTML content'),{
            'fields': ('html_content',),
        }),
        (None,{
            'fields': ('defaultFromName','defaultFromAddress','defaultCC','groupRequired','hideFromForm'),
        }),
    )


######################################
# Staff and subclass admins


class StaffMemberChildAdmin(PolymorphicChildModelAdmin):
    '''
    Base admin class for all child models
    '''
    base_model = StaffMember
    base_fieldsets = (
        (None, {
            'fields': ('firstName','lastName','userAccount',)
        }),
        (_('Contact'), {
            'fields': ('publicEmail','privateEmail','phone'),
        }),
        (_('Bio/Photo'), {
            'fields': ('image','bio'),
        })
    )


@admin.register(Instructor)
class InstructorAdmin(FrontendEditableAdminMixin, StaffMemberChildAdmin):
    base_model = Instructor
    show_in_index = True

    # Allows overriding from other apps
    actions = []
    inlines = []

    list_display = ('fullName','privateEmail','availableForPrivates','status')
    list_display_links = ('fullName',)
    list_editable = ('availableForPrivates','privateEmail','status')
    list_filter = ('status','availableForPrivates')
    search_fields = ('=firstName','=lastName','publicEmail','privateEmail')
    inlines = []

    ordering = ('status','lastName','firstName')

    class Media:
        js = ('bootstrap/js/bootstrap.min.js',)
        css = {'all':('bootstrap/css/bootstrap.min.css',)}


@admin.register(StaffMember)
class StaffMemberParentAdmin(PolymorphicParentModelAdmin):
    '''
    The parent model admin for Staff members
    '''
    base_model = StaffMember
    child_models = (Instructor,)
    list_filter = (PolymorphicChildModelFilter,)

    # Allows overriding from other apps
    inlines = []
    actions = []


######################################
# Event and subclass admins


class EventChildAdmin(PolymorphicChildModelAdmin):
    '''
    Base admin class for all child models
    '''
    base_model = Event

    readonly_fields = ['uuidLink',]

    actions = [repeat_events,]

    def uuidLink(self,obj):
        address = reverse('singleClassRegistration', args=[obj.uuid,])
        return mark_safe('<a href="%s">%s</a>' % (address, address))
    uuidLink.short_description = _('Direct Registration Link')
    uuidLink.allow_tags = True

    # This is needed so that when an event is created, the year and month
    # are properly set right away.
    def save_related(self, request, form, formsets, change):
        obj = form.instance
        super(EventChildAdmin, self).save_related(request, form, formsets, change)

        if obj.eventoccurrence_set.all():
            obj.year = obj.getYearAndMonth()[0]
            obj.month = obj.getYearAndMonth()[1]
            obj.startTime = obj.eventoccurrence_set.order_by('startTime').first().startTime
            obj.endTime = obj.eventoccurrence_set.order_by('endTime').last().endTime
            obj.duration = sum([x.duration for x in obj.eventoccurrence_set.all() if not x.cancelled])
            obj.save()


class SeriesAdminForm(ModelForm):

    def __init__(self, *args, **kwargs):
        super(SeriesAdminForm,self).__init__(*args,**kwargs)

        # Series have registration enabled by default
        self.fields['status'].initial = Event.RegStatus.enabled

        # Locations are required for Series even though they are not for all events.
        self.fields['location'].required = True

        # Allow adding additional rooms from a popup on Location, but not a popup on Room
        self.fields['room'].widget.can_add_related = False
        self.fields['room'].widget.can_change_related = False

        # Impose restrictions on new records, but not on existing ones.
        if not kwargs.get('instance',None):
            # Filter out former locations for new records
            self.fields['location'].queryset = Location.objects.exclude(status=Location.StatusChoices.former)

            # Filter out Pricing Tiers that are expired (i.e. no longer in use)
            self.fields['pricingTier'].queryset = PricingTier.objects.filter(expired=False)

            # Filter out one-time class descriptions
            self.fields['classDescription'].queryset = ClassDescription.objects.filter(oneTimeSeries=False)

            # Set initial values for capacity here because they will automatically update if the
            # constant is changed.  Through Javascript, this should also change when the Location is changed.
            self.fields['capacity'].initial = getConstant(name='registration__defaultEventCapacity')

    # Use the custom location capacity widget to ensure that Javascript can update location specific capacities.
    class Meta:
        model = Series
        exclude = []
        widgets = {
            'location': LocationWithDataWidget,
        }

    class Media:
        js = ('js/serieslocation_capacity_change.js','js/location_related_objects_lookup.js')


@admin.register(Series)
class SeriesAdmin(FrontendEditableAdminMixin, EventChildAdmin):
    base_model = Series
    form = SeriesAdminForm
    show_in_index = True

    inlines = [EventRoleInline,EventOccurrenceInline,SeriesTeacherInline,SubstituteTeacherInline]
    list_display = ('name','series_month','location','class_time','status','registrationOpen','pricingTier','category','session','customers')
    list_editable = ('status','category','session')
    list_filter = ('location','status','registrationOpen','category','session','pricingTier')

    def customers(self,obj):
        return obj.numRegistered
    customers.short_description = _('# Registered Students')

    def series_month(self,obj):
        return '%s %s' % (month_name[obj.month or 0],obj.year or '')

    def class_time(self, obj):
        return obj.startTime.strftime('%A, %I:%M %p')

    fieldsets = (
        (None, {
            'fields': ('classDescription',('location','room'),'pricingTier',('category','session','allowDropins'),('uuidLink',)),
        }),
        (_('Override Display/Registration/Capacity'), {
            'classes': ('collapse',),
            'fields': ('status','closeAfterDays','capacity',),
        }),
    )

    # This allows us to save the obj reference in order to process related objects in an inline (substitute teachers)
    def get_form(self, request, obj=None, **kwargs):
        # just save obj reference for future processing in Inline
        request._obj_ = obj
        return super(SeriesAdmin, self).get_form(request, obj, **kwargs)

    def save_model(self,request,obj,form,change):
        obj.submissionUser = request.user
        obj.save()


class PublicEventAdminForm(ModelForm):

    def __init__(self, *args, **kwargs):
        super(PublicEventAdminForm,self).__init__(*args,**kwargs)

        self.fields['status'].initial = Event.RegStatus.disabled

        # Allow adding additional rooms from a popup on Location, but not a popup on Room
        self.fields['room'].widget.can_add_related = False
        self.fields['room'].widget.can_change_related = False

        # Impose restrictions on new records, but not on existing ones.
        if not kwargs.get('instance',None):
            # Filter out former locations
            self.fields['location'].queryset = Location.objects.exclude(status=Location.StatusChoices.former)

            # Filter out Pricing Tiers that are expired (i.e. no longer in use)
            self.fields['pricingTier'].queryset = PricingTier.objects.filter(expired=False)

            # Set initial values for capacity here because they will automatically update if the
            # constant is changed
            self.fields['capacity'].initial = getConstant(name='registration__defaultEventCapacity')

    # Use the custom location capacity widget to ensure that Javascript can update location specific capacities.
    class Meta:
        model = PublicEvent
        exclude = ['month','year','startTime','endTime','duration','submissionUser','registrationOpen']
        widgets = {
            'location': LocationWithDataWidget,
            'submissionUser': HiddenInput(),
        }

    class Media:
        js = ('js/location_related_objects_lookup.js',)


@admin.register(PublicEvent)
class PublicEventAdmin(FrontendEditableAdminMixin, EventChildAdmin):
    base_model = PublicEvent
    form = PublicEventAdminForm
    show_in_index = True

    list_display = ('name','numOccurrences','firstOccurrenceTime','lastOccurrenceTime','location','status','registrationOpen','pricingTier','category','session','numRegistered')
    list_filter = ('location','status','registrationOpen','pricingTier','category','session')
    list_editable = ('status','category','session')
    search_fields = ('name',)
    ordering = ('-endTime',)
    prepopulated_fields = {'slug': ('title',)}
    inlines = [EventRoleInline,EventOccurrenceInline,EventStaffMemberInline]

    fieldsets = (
        (None, {
            'fields': ('title','slug','category','session',('location','room'),)
        }),
        (_('Registration/Visibility'), {
            'fields': ('status',('pricingTier','capacity'),),
        }),
        (_('Description/Link'), {'fields': ('descriptionField','link')})
    )

    # This allows us to save the obj reference in order to process related objects in an inline (substitute teachers)
    def get_form(self, request, obj=None, **kwargs):
        # just save obj reference for future processing in Inline
        request._obj_ = obj
        return super(PublicEventAdmin, self).get_form(request, obj, **kwargs)

    def save_model(self,request,obj,form,change):
        obj.submissionUser = request.user
        obj.save()


@admin.register(Event)
class EventParentAdmin(PolymorphicParentModelAdmin):
    '''
    The parent model admin for Events
    '''
    list_display = ('name','firstOccurrenceTime','lastOccurrenceTime','location','status','registrationOpen')

    base_model = Event
    child_models = (Series,PublicEvent)
    list_filter = (PolymorphicChildModelFilter,'status','registrationOpen','location')
    list_editable = ('status',)
    polymorphic_list = True

    actions = [repeat_events,]


@admin.register(PublicEventCategory)
class PublicEventCategoryAdmin(admin.ModelAdmin):
    list_display = ['name', 'separateOnRegistrationPage','displayColor']
    prepopulated_fields = {'slug': ('name',)}


@admin.register(SeriesCategory)
class SeriesCategoryAdmin(admin.ModelAdmin):
    list_display = ['name', 'separateOnRegistrationPage']
    prepopulated_fields = {'slug': ('name',)}


# These admin classes are registered but need nothing additional
admin.site.register(DanceRole)
admin.site.register(DanceType)
admin.site.register(DanceTypeLevel)
admin.site.register(EventStaffCategory)



from django.db import models
from django.contrib.auth.models import User
from django.core.exceptions import ValidationError, ObjectDoesNotExist
from django.core.validators import MinValueValidator, MaxValueValidator
from django.utils.translation import ugettext_lazy as _
from django.utils import timezone

from polymorphic.models import PolymorphicModel
from filer.fields.file import FilerFileField
from filer.models import Folder
import math
from djchoices import DjangoChoices, ChoiceItem
from calendar import day_name
from datetime import time, timedelta
from dateutil.relativedelta import relativedelta
from intervaltree import IntervalTree

from danceschool.core.models import StaffMember, EventStaffCategory, Event, InvoiceItem, Location, Room
from danceschool.core.constants import getConstant


def ordinal(n):
    ''' This is just used to populate ordinal day of the month choices '''
    return "%d%s" % (n,"tsnrhtdd"[(math.floor(n / 10) % 10 != 1) * (n % 10 < 4) * n % 10::4])


class ExpenseCategory(models.Model):
    '''
    These are the different available categories of payment
    '''

    name = models.CharField(_('Name'),max_length=50,unique=True,help_text=_('Different types of tasks and payments should have different category names'))
    defaultRate = models.FloatField(_('Default rate'),help_text=_('This is the default hourly payment rate for this type of task.  For staff expenses and venue rentals, this will be overridden by the rate specified as default for the venue or staff type.'),null=True,blank=True,validators=[MinValueValidator(0)])

    def __str__(self):
        return self.name

    class Meta:
        verbose_name = _('Expense category')
        verbose_name_plural = _('Expense categories')


class RevenueCategory(models.Model):
    '''
    These are the different available categories of payment
    '''

    name = models.CharField(_('Name'),max_length=50,unique=True,help_text=_('Different types of revenue fall under different categories.'))
    defaultAmount = models.FloatField(_('Default amount'),help_text=_('This is the default amount of revenue for items in this category.'),null=True,blank=True,validators=[MinValueValidator(0)])

    def __str__(self):
        return self.name

    class Meta:
        verbose_name = _('Revenue category')
        verbose_name_plural = _('Revenue categories')


class RepeatedExpenseRule(PolymorphicModel):
    '''
    This base class defines the pieces of information
    needed for any type of repeated expense creation, such
    as daily/weekly/monthly expense generation for venues
    and instructors, as well as generic repeated expenses.
    '''

    class RateRuleChoices(DjangoChoices):
        hourly = ChoiceItem('H',_('Per hour'))
        daily = ChoiceItem('D',_('Per day of scheduled events'))
        weekly = ChoiceItem('W',_('Per week'))
        monthly = ChoiceItem('M',_('Per month'))

    class MilestoneChoices(DjangoChoices):
        start = ChoiceItem('S',_('First occurrence starts'))
        end = ChoiceItem('E',_('Last occurrence ends'))

    rentalRate = models.FloatField(
        _('Expense Rate'),validators=[MinValueValidator(0)],help_text=_('In default currency')
    )

    applyRateRule = models.CharField(
        _('Apply this rate'),
        max_length=1,
        choices=RateRuleChoices.choices,
        default=RateRuleChoices.hourly,
    )

    dayStarts = models.PositiveSmallIntegerField(
        _('Day starts at'),
        choices=[(i, time(i).strftime('%-I:00 %p')) for i in range(24)],
        default=0,
        validators=[MinValueValidator(0),MaxValueValidator(23)],
        help_text=_('If you run events after midnight, avoids creation of duplicate expense items'),
    )

    weekStarts = models.PositiveSmallIntegerField(
        _('Week starts on'),
        choices=[(x,day_name[x]) for x in range(0,7)],
        default=0,
        validators=[MinValueValidator(0),MaxValueValidator(6)]
    )

    monthStarts = models.PositiveSmallIntegerField(
        _('Month starts on'),
        choices=[(x, ordinal(x)) for x in range(1,29)],
        default=1,
        validators=[MinValueValidator(1),MaxValueValidator(28)]
    )

    startDate = models.DateField(
        _('Start date'),
        null=True,
        blank=True,
        help_text=_('If specified, then expense items will not be generated prior to this date.')
    )

    endDate = models.DateField(
        _('End date'),
        null=True,
        blank=True,
        help_text=_('If specified, then expense items will not be generated after this date.  Leave blank for expenses to be generated indefinitely.')
    )

    advanceDays = models.PositiveSmallIntegerField(
        _('Generate expenses up to __ days in advance'),
        default=30,
    )

    advanceDaysReference = models.CharField(
        _('in advance of'),
        max_length=1,
        choices=MilestoneChoices.choices,
        default=MilestoneChoices.start,
    )

    priorDays = models.PositiveSmallIntegerField(
        _('Generate expenses up to __ days in the past'),
        help_text=_('Leave blank for no limit.'),
        default=180,
        null=True,
        blank=True
    )

    priorDaysReference = models.CharField(
        _('prior to'),
        max_length=1,
        choices=MilestoneChoices.choices,
        default=MilestoneChoices.end,
    )

    disabled = models.BooleanField(
        _('Disable autogeneration of expenses under this rule'),
        default=False,
        help_text=_(
            'It is recommended to disable expense rules rather than delete them for ' +
            'temporary purposes, to avoid the creation of duplicate expense items.'
        ),
    )

    lastRun = models.DateTimeField(_('Last run time'),null=True,blank=True)

    def generateExpenses(self,request=None,datetimeTuple=None):
        '''
        Child classes that define this method can have their expense rule run
        from the admin interface.
        '''
        pass

    def timeAtThreshold(self,dateTime):
        '''
        A convenience method for checking when a time is on the start/end boundary
        for this rule.
        '''

        # Anything that's not at a day boundary is for sure not at a threshold.
        if not (
            dateTime.hour == self.dayStarts and dateTime.minute == 0 and
            dateTime.second == 0 and dateTime.microsecond == 0
        ):
            return False

        if self.applyRateRule == self.RateRuleChoices.daily:
            return True
        elif self.applyRateRule == self.RateRuleChoices.weekly:
            return (dateTime == self.weekStarts)
        elif self.applyRateRule == self.RateRuleChoices.monthly:
            return (dateTime.day == self.monthStarts)

        # Everything else is nonsensical, so False.
        return False

    def getWindowsAndTotals(self,intervals):

        # Ensure that the intervals are passed with startTime and endTime in order, and reduce
        # the intervals down to non-overlapping intervals.
        intervals = [sorted(x) for x in intervals]
        tree = IntervalTree.from_tuples(intervals)
        tree.merge_overlaps()

        # This is the set of times at which weekly or monthly intervals need to be sliced
        # (i.e. the start of each new week or month period).
        slice_times = set()

        # Using the set of passed intervals, construct the set of day,
        # week, or monthly intervals specified by the expense rule.
        for startTime, endTime in [(x.begin, x.end) for x in tree]:

            if self.applyRateRule == self.RateRuleChoices.daily:
                # Period is the date or dates of the occurrence
                this_window_start = startTime.replace(hour=self.dayStarts,minute=0,second=0,microsecond=0)
                this_window_end = (endTime + timedelta(days=1)).replace(hour=self.dayStarts,minute=0,second=0,microsecond=0) \
                    if not self.timeAtThreshold(endTime) else endTime
            elif self.applyRateRule == self.RateRuleChoices.weekly:
                # Period is the week of the occurrence, starting from the start date specified for
                # the Location.
                if startTime.weekday() > self.weekStarts:
                    start_offset = self.weekStarts - startTime.weekday()
                else:
                    start_offset = self.weekStarts - startTime.weekday() - 7

                if endTime.weekday() > self.weekStarts or (endTime.weekday() == self.weekStarts and not self.timeAtThreshold(endTime)):
                    end_offset = 7 + self.weekStarts - endTime.weekday()
                else:
                    end_offset = self.weekStarts - endTime.weekday()

                this_window_start = (startTime + timedelta(days=start_offset)).replace(hour=self.dayStarts, minute=0, second=0, microsecond=0)
                this_window_end = (endTime + timedelta(days=end_offset)).replace(hour=self.dayStarts, minute=0, second=0, microsecond=0)

                # Add the weekly slice times to the set
                t0 = this_window_start
                while t0 <= this_window_end:
                    slice_times.add(t0)
                    t0 = t0 + timedelta(days=7)

            elif self.applyRateRule == self.RateRuleChoices.monthly:
                # Period is the month of the occurrence, starting from the start date specified for
                # the Location.
                startDay = self.monthStarts

                if startTime.day >= startDay:
                    this_window_start = startTime.replace(day=startDay,hour=self.dayStarts,minute=0,second=0,microsecond=0)
                else:
                    this_window_start = (startTime + relativedelta(months=-1)).replace(day=startDay,hour=self.dayStarts,minute=0,second=0,microsecond=0)
                if endTime.day > startDay or (endTime.day == startDay and not self.timeAtThreshold(endTime)):
                    this_window_end = (endTime + relativedelta(months=1)).replace(day=startDay,hour=self.dayStarts,minute=0,second=0,microsecond=0)
                else:
                    this_window_end = endTime.replace(day=startDay,hour=self.dayStarts,minute=0,second=0,microsecond=0)

                # Add the monthly slice times to the set
                t0 = this_window_start
                while t0 <= this_window_end:
                    slice_times.add(t0)
                    t0 = t0 + relativedelta(months=1)

            # Add the newly constructed interval to the tree of intervals
            tree.addi(this_window_start, this_window_end)

        # Remove all overlapping intervals and also chop out any intervals for which there is already an
        # expense item existing.  (we will split again by week or month afterward.)
        tree.merge_overlaps()

        startTime = tree.begin()
        endTime = tree.end()

        overlapping = self.expenseitem_set.filter(
            (models.Q(periodStart__lte=endTime) & models.Q(periodStart__gte=startTime)) |
            (models.Q(periodEnd__gte=startTime) & models.Q(periodEnd__lte=endTime)) |
            (models.Q(periodStart__lte=startTime) & models.Q(periodEnd__gte=endTime))
        )

        for item in overlapping:
            tree.chop(item.periodStart, item.periodEnd)

        # Now merge the intervals again, and split at the split times if needed
        tree.merge_overlaps()
        for slice_time in slice_times:
            tree.slice(slice_time)

        # Now, loop through the items of the finalized tree and yield the times,
        # a description and total expense for the interval allocated by the fraction
        # of a full week/month interval contained in the interval so that new
        # ExpenseItems may be created.
        for startTime, endTime in [(x.begin, x.end) for x in tree]:

            # Default description is overridden below as appropriate
            description = str(_('%(start)s to %(end)s' % {
                'start': startTime.strftime('%Y-%m-%d'),
                'end': (endTime - timedelta(hours=self.dayStarts,minutes=1)).strftime('%Y-%m-%d')
            }))

            if self.applyRateRule == self.RateRuleChoices.daily:
                num_days = (endTime - startTime).days
                total = self.rentalRate * num_days

                if num_days == 1:
                    description = startTime.strftime('%Y-%m-%d')

            elif self.applyRateRule == self.RateRuleChoices.weekly:
                num_days = (endTime - startTime).days
                total = self.rentalRate * (num_days / 7)

                if num_days == 7:
                    description = str(_('week of %(start)s to %(end)s' % {
                        'start': startTime.strftime('%Y-%m-%d'),
                        'end': (endTime - timedelta(hours=self.dayStarts,minutes=1)).strftime('%Y-%m-%d')
                    }))

            elif self.applyRateRule == self.RateRuleChoices.monthly:
                num_days = (endTime - startTime).days

                # We need to know the number days in the month in order to allocate partial expenses
                month_startDate = startTime.replace(day=startDay,hour=self.dayStarts,minute=0,second=0,microsecond=0)
                month_startDate = month_startDate - relativedelta(months=1) if month_startDate > startTime else month_startDate
                days_in_month = (month_startDate + relativedelta(months=1) - month_startDate).days
                total = self.rentalRate * (num_days / days_in_month)

                if num_days == days_in_month:
                    description = str(_('month of %(start)s to %(end)s' % {
                        'start': startTime.strftime('%Y-%m-%d'),
                        'end': (endTime - timedelta(hours=self.dayStarts,minutes=1)).strftime('%Y-%m-%d')
                    }))

            # Yield the information for this interval
            yield (startTime, endTime, total, description)

    @property
    def ruleName(self):
        ''' This should be overridden for child classes '''
        return '%s %s' % (self.rentalRate, self.RateRuleChoices.values.get(self.applyRateRule,self.applyRateRule))
    ruleName.fget.short_description = _('Rule name')

    def __str__(self):
        ''' Should be overridden by child classes with something more descriptive. '''
        return str(_('Repeated expense rule: %s' % self.name))


class LocationRentalInfo(RepeatedExpenseRule):
    '''
    This model is used to store information on rental periods and rates
    for locations.
    '''
    location = models.OneToOneField(Location,related_name='rentalinfo',verbose_name=_('Location'))

    @property
    def ruleName(self):
        ''' overrides from parent class '''
        return self.location.name
    ruleName.fget.short_description = _('Rule name')

    def generateExpenses(self,request=None,datetimeTuple=None):
        from .helpers import createExpenseItemsForVenueRental
        return createExpenseItemsForVenueRental(rule=self,request=request,datetimeTuple=datetimeTuple)

    def __str__(self):
        return str(_('Rental expense information for: %s' % self.location.name))

    class Meta:
        verbose_name = _('Location rental information')
        verbose_name_plural = _('Locations\' rental information')


class RoomRentalInfo(RepeatedExpenseRule):
    '''
    This model is used to store information on rental periods and rates
    for individual rooms.  If a rental rate does not exist for a room,
    or if it is specified that the room rental rate not be applied,
    then the location's rental rate and parameters are used instead.
    '''
    room = models.OneToOneField(Room,related_name='rentalinfo',verbose_name=_('Room'))

    @property
    def ruleName(self):
        ''' overrides from parent class '''
        return _('%s at %s' % (self.room.name, self.room.location.name))
    ruleName.fget.short_description = _('Rule name')

    def generateExpenses(self,request=None,datetimeTuple=None):
        from .helpers import createExpenseItemsForVenueRental
        return createExpenseItemsForVenueRental(rule=self,request=request,datetimeTuple=datetimeTuple)

    def __str__(self):
        return str(_('Rental expense information for: %s at %s' % (self.room.name, self.room.location.name)))

    class Meta:
        verbose_name = _('Room rental information')
        verbose_name_plural = _('Rooms\' rental information')


class StaffMemberWageInfo(RepeatedExpenseRule):
    '''
    This model is used to store information on rental periods and rates
    for individual rooms.  If a rental rate does not exist for a room,
    or if it is specified that the room rental rate not be applied,
    then the location's rental rate and parameters are used instead.
    '''
    staffMember = models.ForeignKey(StaffMember,related_name='expenserules',verbose_name=_('Staff member'))
    category = models.ForeignKey(
        EventStaffCategory,
        verbose_name=_('Category'),
        null=True,blank=True,
        help_text=_('If left blank, then this expense rule will be used for all categories.  If a category-specific rate is specified, then that will be used instead.  If nothing is specified for an instructor, then the default hourly rate for each category will be used.'),
        on_delete=models.SET_NULL,
    )

    @property
    def ruleName(self):
        ''' overrides from parent class '''
        return _('%s: %s' % (self.staffMember.fullName, self.category or 'All unspecified categories'))
    ruleName.fget.short_description = _('Rule name')

    def generateExpenses(self,request=None,datetimeTuple=None):
        from .helpers import createExpenseItemsForEvents
        return createExpenseItemsForEvents(rule=self,request=request,datetimeTuple=datetimeTuple)

    def __str__(self):
        return str(_('%s wage/salary information for: %s' % (self.category or 'Rental', self.staffMember.fullName)))

    class Meta:
        unique_together = ('staffMember', 'category')
        ordering = ('staffMember','category__name')
        verbose_name = _('Staff member salary information')
        verbose_name_plural = _('Staff members\' wage/salary information')


class GenericRepeatedExpense(RepeatedExpenseRule):
    '''
    This model is used to store repeated expenses that are not specifically tied to location
    rental or event staffing.  That is, expenses are generated under these rules regardless
    of whether any series or events are booked.
    '''

    name = models.CharField(_('Give this expense generation rule a name'),max_length=100,unique=True)

    category = models.ForeignKey(ExpenseCategory,verbose_name=_('Category'),null=True,on_delete=models.SET_NULL,)

    # An expense can be directly associated with a user (like an instructor) or
    # a location, or the name of another party can be entered.
    payToUser = models.ForeignKey(
        User,null=True,blank=True,
        verbose_name=_('Pay to user'),on_delete=models.SET_NULL,)
    payToLocation = models.ForeignKey(Location,null=True,blank=True,verbose_name=_('Pay to location'))
    payToName = models.CharField(_('Pay to (enter name)'),max_length=50, null=True,blank=True)

    markApproved = models.BooleanField(_('Automatically mark this expense as approved'),default=False)
    markPaid = models.BooleanField(_('Automatically mark this expense as paid'),default=False)
    paymentMethod = models.CharField(
        _('Payment method'),
        max_length=50,null=True,blank=True,
        help_text=_('This field is ignored unless you have chosen to automatically mark expenses as paid.')
    )

    @property
    def ruleName(self):
        ''' overrides from parent class '''
        return self.name
    ruleName.fget.short_description = _('Rule name')

    def clean(self):
        ''' priorDays is required for Generic Repeated Expenses to avoid infinite loops '''
        if not self.priorDays and not self.startDate:
            raise ValidationError(_(
                'Either a start date or an "up to __ days in the past" limit is required ' +
                'for repeated expense rules that are not associated with a venue or a staff member.'
            ))
        super(GenericRepeatedExpense,self).clean()

    def generateExpenses(self,request=None,datetimeTuple=None):
        from .helpers import createGenericExpenseItems
        return createGenericExpenseItems(rule=self,request=request,datetimeTuple=datetimeTuple)

    def __str__(self):
        return str(_('Repeated expense rule: %s' % self.name))

    class Meta:
        ordering = ('name',)
        verbose_name = _('Other repeated expense')
        verbose_name_plural = _('Other repeated expenses')


class ExpenseItem(models.Model):
    '''
    Expenses may be associated with EventStaff or with Events, or they may be associated with nothing
    '''

    submissionUser = models.ForeignKey(
        User,
        verbose_name=_('Submission user'),
        related_name='expensessubmittedby',
        null=True,blank=True,on_delete=models.SET_NULL,)
    submissionDate = models.DateTimeField(_('Submission date'),auto_now_add=True)

    category = models.ForeignKey(ExpenseCategory,verbose_name=_('Category'),null=True,on_delete=models.SET_NULL,)

    description = models.CharField(_('Description'),max_length=200,null=True,blank=True)

    hours = models.FloatField(
        _('Hours'),
        help_text=_('Please indicate the number of hours to be paid for.'),
        null=True,blank=True,validators=[MinValueValidator(0)])
    wageRate = models.FloatField(
        _('Wage rate'),
        help_text=_('This should be filled automatically, but can be changed as needed.'),
        null=True,blank=True,validators=[MinValueValidator(0)])
    total = models.FloatField(_('Total amount'),null=True,blank=True,validators=[MinValueValidator(0)])
    adjustments = models.FloatField(
        _('Adjustments/refunds'),
        help_text=_('Record any ex-post adjustments to the amount (e.g. refunds) in this field. A positive amount increases the netExpense, a negative amount reduces the netExpense.'),
        default=0)
    fees = models.FloatField(
        _('Fees'),
        help_text=_('The sum of any transaction fees (e.g. Paypal fees) that were paid <strong>by us</strong>, and should therefore be added to net expense.'),
        default=0)

    paymentMethod = models.CharField(_('Payment method'),max_length=50,null=True,blank=True)

    comments = models.TextField(_('Comments/Notes'),null=True,blank=True)
    attachment = FilerFileField(verbose_name=_('Attach File (optional)'),null=True,blank=True,related_name='expense_attachment')

    # An expense item will automatically be associated with an event if it is an automatically-generated
    # _hourly_ expense for venue rental or for event staff.  This facilitates event-level financial statements.
    # Non-hourly generated expense items are not automatically affiliated with events.  It is also possible
    # to affiliate an Expense Item with an event via the Expense Reporting Form.
    event = models.ForeignKey(
        Event,
        null=True,blank=True,
        verbose_name=_('Event'),
        help_text=_('If this item is associated with an Event, enter it here.'),on_delete=models.SET_NULL,)

    # For periodic expenses (e.g. hourly/daily/weekly/monthly venue rental, instructor expenses, etc.  This foreign key
    # also replaces the prior relations eventstaffmember and eventvenue, because all automatically generated expenses
    # are now generated against specific repeated expense rules.
    expenseRule = models.ForeignKey(
        RepeatedExpenseRule,
        verbose_name=_('Expense generation rule'),
        null=True,blank=True,on_delete=models.SET_NULL,)

    # For daily/weekly/monthly automatically-generate expenses, this defines the period over which
    # this expense item applies.
    periodStart = models.DateTimeField(_('Expense period start'),null=True,blank=True)
    periodEnd = models.DateTimeField(_('Expense period end'),null=True,blank=True)

    # An expense can be directly associated with a user (like an instructor) or
    # a location, or the name of another party can be entered.
    payToUser = models.ForeignKey(
        User,null=True,blank=True,
        related_name='payToUser',verbose_name=_('Pay to user'),on_delete=models.SET_NULL,)
    payToLocation = models.ForeignKey(Location,null=True,blank=True,verbose_name=_('Pay to location'))
    payToName = models.CharField(_('Pay to (enter name)'),max_length=50, null=True,blank=True)

    reimbursement = models.BooleanField(
        _('Reimbursement'),
        help_text=_('Check to indicate that this is a reimbursement expense (i.e. not compensation).'),
        default=False)

    approved = models.BooleanField(
        _('Approved'),
        help_text=_('Check to indicate that expense is approved for payment.'),
        default=False)
    paid = models.BooleanField(
        _('Paid'),
        help_text=_('Check to indicate that payment has been made.'),
        default=False)

    approvalDate = models.DateTimeField(_('Approval date'),null=True,blank=True)
    paymentDate = models.DateTimeField(_('Payment date'),null=True,blank=True)

    # This field is used to aggregate expenses over time (e.g. by month).
    # The value of this field is auto-updated using pre-save methods. If
    # there is a class series or an event associated with this expense,
    # then the value is taken from that.  Otherwise, the submission date
    # is used.
    accrualDate = models.DateTimeField(_('Accrual date'))

    @property
    def netExpense(self):
        return self.total + self.adjustments + self.fees
    netExpense.fget.short_description = _('Net expense')

    @property
    def expenseStartDate(self):
        theTime = self.accrualDate
        if self.periodStart:
            theTime = self.periodStart
        elif self.event:
            theTime = self.event.startTime
        return theTime
    expenseStartDate.fget.short_description = _('Start Date')

    @property
    def expenseEndDate(self):
        theTime = self.accrualDate
        if self.periodEnd:
            theTime = self.periodEnd
        elif self.event:
            theTime = self.event.endTime
        return theTime
    expenseEndDate.fget.short_description = _('End Date')

    @property
    def payTo(self):
        '''
        Returns a string indicating who the expense is to be paid to.
        For more convenient references of miscellaneous expenses.
        '''
        if self.payToUser:
            return ' '.join([self.payToUser.first_name,self.payToUser.last_name])
        elif self.payToLocation:
            return self.payToLocation.name
        else:
            return self.payToName or ''
    payTo.fget.short_description = _('Pay to')

    def save(self, *args, **kwargs):
        '''
        This custom save method ensures that an expense is not attributed to multiple categories.
        It also ensures that the series and event properties are always associated with any
        type of expense of that series or event.
        '''
        # Set the approval and payment dates if they have just been approved/paid.
        if not hasattr(self,'__paid') or not hasattr(self,'__approved'):
            if self.approved and not self.approvalDate:
                self.approvalDate = timezone.now()
            if self.paid and not self.paymentDate:
                self.paymentDate = timezone.now()
        else:
            if self.approved and not self.approvalDate and not self.__approvalDate:
                self.approvalDate = timezone.now()
            if self.paid and not self.paymentDate and not self.__paymentDate:
                self.paymentDate = timezone.now()

        # Fill out the series and event properties to permit easy calculation of
        # revenues and expenses by series or by event.
        if (
            self.expenseRule and hasattr(self.expenseRule,'staffMember') and
            hasattr(self.expenseRule.staffMember,'userAccount') and not self.payToUser
        ):
            self.payToUser = self.expenseRule.staffMember.userAccount
        if (
            self.expenseRule and hasattr(self.expenseRule,'location') and
            not self.payToLocation
        ):
            self.payToLocation = self.expenseRule.location

        # Set the accrual date.  The method for events ensures that the accrualDate month
        # is the same as the reported month of the series/event by accruing to the end date of the last
        # class or occurrence in that month.
        if not self.accrualDate:
            if self.event and self.event.month:
                self.accrualDate = self.event.eventoccurrence_set.order_by('endTime').filter(**{'endTime__month': self.event.month}).last().endTime
            elif self.submissionDate:
                self.accrualDate = self.submissionDate
            else:
                self.submissionDate = timezone.now()
                self.accrualDate = self.submissionDate

        # Set the total for hourly work
        if self.hours and not self.wageRate and not self.total and not self.payToLocation and self.category:
            self.wageRate = self.category.defaultRate
        elif self.hours and not self.wageRate and not self.total and self.payToLocation:
            self.wageRate = self.payToLocation.rentalRate

        if self.hours and self.wageRate and not self.total:
            self.total = self.hours * self.wageRate

        super(ExpenseItem, self).save(*args, **kwargs)
        self.__approved = self.approved
        self.__paid = self.paid
        self.__approvalDate = self.approvalDate
        self.__paymentDate = self.paymentDate

        # If a file is attached, ensure that it is not public, and that it is saved in the 'Expense Receipts' folder
        if self.attachment:
            try:
                self.attachment.folder = Folder.objects.get(name=_('Expense Receipts'))
            except ObjectDoesNotExist:
                pass
            self.attachment.is_public = False
            self.attachment.save()

    def __str__(self):
        if self.accrualDate:
            return '%s %s: %s = %s%s' % (self.category.name, self.accrualDate.strftime('%B %Y'),self.description, getConstant('general__currencySymbol'), self.total)
        else:
            return '%s: %s = %s%s' % (self.category.name, self.description, getConstant('general__currencySymbol'), self.total)

    def __init__(self, *args, **kwargs):
        '''
        Permit easy checking to determine if the object
        already exists and has changed on saving
        '''
        super(self.__class__, self).__init__(*args, **kwargs)
        self.__approved = self.approved
        self.__paid = self.paid
        self.__approvalDate = self.approvalDate
        self.__paymentDate = self.paymentDate

    class Meta:
        ordering = ['-accrualDate',]
        verbose_name = _('Expense item')
        verbose_name_plural = _('Expense items')

        permissions = (
            ('mark_expenses_paid',_('Mark expenses as paid at the time of submission')),
        )


class RevenueItem(models.Model):
    '''
    All revenue-producing transactions (e.g. class payments, other payments) should have an associated RevenueItem
    '''

    submissionUser = models.ForeignKey(
        User,null=True,blank=True,related_name='revenuessubmittedby',
        verbose_name=_('Submission user'),on_delete=models.SET_NULL,)
    submissionDate = models.DateTimeField(_('Submission date'),auto_now_add=True)

    category = models.ForeignKey(RevenueCategory,verbose_name=_('Category'),null=True,on_delete=models.SET_NULL,)
    description = models.CharField(_('Description'),max_length=200,null=True,blank=True)
    total = models.FloatField(_('Total'),help_text=_('The total revenue received, net of any discounts or voucher uses.  This is what we actually receive.'),validators=[MinValueValidator(0)])
    grossTotal = models.FloatField(_('Gross Total'),help_text=_('The gross total billed before the application of any discounts, or the use of any vouchers.'),validators=[MinValueValidator(0)])
    adjustments = models.FloatField(_('Adjustments'),help_text=_('Record any ex-post adjustments to the amount (e.g. refunds) in this field.  A positive amount increases the netRevenue, a negative amount reduces the netRevenue.'),default=0)
    fees = models.FloatField(_('Fees'),help_text=_('The sum of any transaction fees (e.g. Paypal fees) that were paid <strong>by us</strong>, and should therefore be subtracted from net revenue.'),default=0)
    taxes = models.FloatField(_('Taxes'),default=0)
    buyerPaysSalesTax = models.BooleanField(_('Buyer pays sales tax'), default=False)

    paymentMethod = models.CharField(_('Payment method'),max_length=50,null=True,blank=True)
    invoiceNumber = models.CharField(_('Invoice Number'),help_text=_('For Paypal payments, this will be the txn_id.  For cash payments, this will be automatically generated by the submission form.  More than one revenue item may have the same invoice number, because multiple events are paid for in one Paypal transaction.'),null=True,blank=True,max_length=80)

    comments = models.TextField(_('Comments/Notes'),null=True,blank=True)
    attachment = FilerFileField(verbose_name=_('Attach File (optional)'),null=True,blank=True,related_name='revenue_attachment')

    # With the invoice system in the core app, Revenue Items need only link with Invoice Items
    invoiceItem = models.OneToOneField(InvoiceItem,null=True, blank=True,verbose_name=_('Associated invoice item'))

    event = models.ForeignKey(Event,null=True,blank=True,verbose_name=_('Event'),help_text=_('If this item is associated with an Event, enter it here.'))
    receivedFromName = models.CharField(_('Received From'),max_length=50,null=True,blank=True,help_text=_('Enter who this revenue item was received from, if it is not associated with an existing registration.'))

    currentlyHeldBy = models.ForeignKey(User,null=True,blank=True,verbose_name=_('Cash currently in possession of'),help_text=_('If cash has not yet been deposited, this indicates who to contact in order to collect the cash for deposit.'),related_name='revenuesheldby')
    received = models.BooleanField(_('Received'),help_text=_('Check to indicate that payment has been received. Non-received payments are considered pending.'),default=False)
    receivedDate = models.DateTimeField(_('Date received'),null=True,blank=True)

    # This field is used to aggregate expenses over time (e.g. by month).
    # The value of this field is auto-updated using pre-save methods. If
    # there is a registration or an event associated with this expense,
    # then the value is taken from that.  Otherwise, the submission date
    # is used.
    accrualDate = models.DateTimeField(_('Accrual date'))

    @property
    def relatedItems(self):
        '''
        If this item is associated with a registration, then return all other items associated with
        the same registration.
        '''
        if self.registration:
            return self.registration.revenueitem_set.exclude(pk=self.pk)
    relatedItems.fget.short_description = _('Related items')

    @property
    def netRevenue(self):
        net = self.total + self.adjustments - self.fees
        if not self.buyerPaysSalesTax:
            net -= self.taxes
        return net
    netRevenue.fget.short_description = _('Net revenue')

    def save(self, *args, **kwargs):
        '''
        This custom save method ensures that a revenue item is not attributed to multiple categories.
        It also ensures that the series and event properties are always associated with any
        type of revenue of that series or event.
        '''

        # Set the received date if the payment was just marked received
        if not hasattr(self,'__received'):
            if self.received and not self.receivedDate:
                self.receivedDate = timezone.now()
        else:
            if self.received and not self.receivedDate and not self.__receivedDate:
                self.receivedDate = timezone.now()

        # Set the accrual date.  The method for series/events ensures that the accrualDate month
        # is the same as the reported month of the event/series by accruing to the start date of the first
        # occurrence in that month.
        if not self.accrualDate:
            if self.invoiceItem and self.invoiceItem.finalEventRegistration:
                min_event_time = self.invoiceItem.finalEventRegistration.event.eventoccurrence_set.filter(**{'startTime__month':self.invoiceItem.finalEventRegistration.event.month}).first().startTime
                self.accrualDate = min_event_time
            elif self.event:
                self.accrualDate = self.event.eventoccurrence_set.order_by('startTime').filter(**{'startTime__month': self.event.month}).last().startTime
            elif self.invoiceItem:
                self.accrualDate = self.invoiceItem.invoice.creationDate
            elif self.receivedDate:
                self.accrualDate = self.receivedDate
            elif self.submissionDate:
                self.accrualDate = self.submissionDate
            else:
                self.submissionDate = timezone.now()
                self.accrualDate = self.submissionDate

        # Now, set the registration property and check that this item is not attributed
        # to multiple categories.
        if self.invoiceItem and self.invoiceItem.finalEventRegistration:
            self.event = self.invoiceItem.finalEventRegistration.event
        elif self.invoiceItem and self.invoiceItem.temporaryEventRegistration:
            self.event = self.invoiceItem.temporaryEventRegistration.event

        # If no grossTotal is reported, use the net total.  If no net total is reported, use the grossTotal
        if self.grossTotal is None and self.total:
            self.grossTotal = self.total
        if self.total is None and self.grossTotal:
            self.total = self.grossTotal

        super(RevenueItem, self).save(*args, **kwargs)
        self.__received = self.received
        self.__receivedDate = self.receivedDate

        # If a file is attached, ensure that it is not public, and that it is saved in the 'Expense Receipts' folder
        if self.attachment:
            try:
                self.attachment.folder = Folder.objects.get(name=_('Revenue Receipts'))
            except ObjectDoesNotExist:
                pass
            self.attachment.is_public = False
            self.attachment.save()

    def __str__(self):
        if self.accrualDate:
            return '%s %s: %s = %s%s' % (self.category.name, self.accrualDate.strftime('%B %Y'),self.description, getConstant('general__currencySymbol'), self.total)
        else:
            return '%s: %s = %s%s' % (self.category.name, self.description, getConstant('general__currencySymbol'), self.total)

    def __init__(self,*args,**kwargs):
        '''
        Permit easy checking to determine if the object
        already exists and has changed on saving
        '''
        super(self.__class__, self).__init__(*args, **kwargs)
        self.__received = self.received
        self.__receivedDate = self.receivedDate

    class Meta:
        ordering = ['-accrualDate',]
        verbose_name = _('Revenue item')
        verbose_name_plural = _('Revenue items')

        permissions = (
            ('export_financial_data',_('Export detailed financial transaction information to CSV')),
            ('view_finances_bymonth',_('View school finances month-by-month')),
            ('view_finances_byevent',_('View school finances by Event')),
            ('view_finances_detail',_('View school finances as detailed statement')),
        )



from django.contrib import admin
from django.contrib.auth.models import User
from django.contrib.contenttypes.models import ContentType
from django.core.urlresolvers import reverse
from django.db.models import Q
from django.utils.html import format_html
from django.forms import ModelForm, ModelChoiceField
from django.utils.translation import ugettext_lazy as _
from django.http import HttpResponseRedirect

from dal import autocomplete
from daterange_filter.filter import DateRangeFilter
from polymorphic.admin import PolymorphicParentModelAdmin, PolymorphicChildModelAdmin, PolymorphicChildModelFilter

from danceschool.core.models import Location, Room, StaffMember, Instructor

from .models import ExpenseItem, ExpenseCategory, RevenueItem, RevenueCategory, RepeatedExpenseRule, LocationRentalInfo, RoomRentalInfo, StaffMemberWageInfo, GenericRepeatedExpense
from .forms import ExpenseCategoryWidget
from .autocomplete_light_registry import get_method_list


class ExpenseItemAdminForm(ModelForm):
    payToUser = ModelChoiceField(
        queryset=User.objects.all(),
        label=_('Pay to this user'),
        required=False,
        widget=autocomplete.ModelSelect2(
            url='autocompleteUser',
            attrs={
                # This will set the input placeholder attribute:
                'data-placeholder': _('Enter a user name'),
                # This will set the yourlabs.Autocomplete.minimumCharacters
                # options, the naming conversion is handled by jQuery
                'data-minimum-input-length': 2,
                'data-max-results': 4,
                'class': 'modern-style',
            }
        )
    )
    paymentMethod = autocomplete.Select2ListCreateChoiceField(
        choice_list=get_method_list,
        required=False,
        widget=autocomplete.ListSelect2(url='paymentMethod-list-autocomplete')
    )

    class Meta:
        model = ExpenseItem
        exclude = []
        widgets = {
            'category': ExpenseCategoryWidget
        }


@admin.register(ExpenseItem)
class ExpenseItemAdmin(admin.ModelAdmin):
    form = ExpenseItemAdminForm

    list_display = ('category','expenseStartDate','expenseEndDate','description','hours','total','approved','paid','reimbursement','payTo','paymentMethod')
    list_editable = ('approved','paid','paymentMethod')
    search_fields = ('description','comments','=category__name','=payToUser__first_name','=payToUser__last_name','=payToLocation__name')
    list_filter = ('category','approved','paid','paymentMethod','reimbursement','payToLocation',('accrualDate',DateRangeFilter),('paymentDate',DateRangeFilter),('submissionDate',DateRangeFilter),'expenseRule')
    readonly_fields = ('submissionUser','expenseRule','expenseStartDate','expenseEndDate')
    actions = ('approveExpense','unapproveExpense')

    fieldsets = (
        (_('Basic Info'), {
            'fields': ('category','description','hours','wageRate','total','adjustments','fees','reimbursement','comments')
        }),
        (_('File Attachment (optional)'), {
            'fields': ('attachment',)
        }),
        (_('Approval/Payment Status'), {
            'classes': ('collapse',),
            'fields': ('periodStart','periodEnd','approved','approvalDate','paid','paymentDate','paymentMethod','accrualDate','expenseRule','event','payToUser','payToLocation','payToName')
        }),
    )

    def approveExpense(self, request, queryset):
        rows_updated = queryset.update(approved=True)
        if rows_updated == 1:
            message_bit = "1 expense item was"
        else:
            message_bit = "%s expense items were" % rows_updated
        self.message_user(request, "%s successfully marked as approved." % message_bit)
    approveExpense.short_description = _('Mark Expense Items as approved')

    def unapproveExpense(self, request, queryset):
        rows_updated = queryset.update(approved=False)
        if rows_updated == 1:
            message_bit = "1 expense item was"
        else:
            message_bit = "%s expense items were" % rows_updated
        self.message_user(request, "%s successfully marked as not approved." % message_bit)
    unapproveExpense.short_description = _('Mark Expense Items as not approved')

    def get_changelist_form(self, request, **kwargs):
        ''' Ensures that the autocomplete view works for payment methods. '''
        return ExpenseItemAdminForm

    def save_model(self,request,obj,form,change):
        obj.submissionUser = request.user
        obj.save()

    class Media:
        js = ('js/update_task_wages.js',)


class RevenueItemAdminForm(ModelForm):
    currentlyHeldBy = ModelChoiceField(
        queryset=User.objects.filter(Q(staffmember__isnull=False) | Q(is_staff=True)),
        label=_('Cash currently in possession of'),
        required=False,
        widget=autocomplete.ModelSelect2(
            url='autocompleteUser',
            attrs={
                # This will set the input placeholder attribute:
                'data-placeholder': _('Enter a user name'),
                # This will set the yourlabs.Autocomplete.minimumCharacters
                # options, the naming conversion is handled by jQuery
                'data-minimum-input-length': 2,
                'data-max-results': 4,
                'class': 'modern-style',
            }
        )
    )
    paymentMethod = autocomplete.Select2ListChoiceField(
        choice_list=get_method_list,
        required=False,
        widget=autocomplete.ListSelect2(url='paymentMethod-list-autocomplete')
    )

    class Meta:
        model = RevenueItem
        exclude = []


@admin.register(RevenueItem)
class RevenueItemAdmin(admin.ModelAdmin):
    form = RevenueItemAdminForm

    list_display = ('description','category','grossTotal','total','adjustments','netRevenue','received','receivedDate','invoiceLink')
    list_editable = ('received',)
    search_fields = ('description','comments','invoiceItem__id','invoiceItem__invoice__id')
    list_filter = ('category','received','paymentMethod',('receivedDate',DateRangeFilter),('accrualDate',DateRangeFilter),('submissionDate',DateRangeFilter))
    readonly_fields = ('netRevenue','submissionUserLink','relatedRevItemsLink','eventLink','paymentMethod','invoiceNumber','invoiceLink')
    actions = ('markReceived','markNotReceived')

    fieldsets = (
        (_('Basic Info'), {
            'fields': ('category','description','grossTotal','total','adjustments','fees','netRevenue','paymentMethod','receivedFromName','invoiceNumber','comments')
        }),
        (_('Related Items'),{
            'fields': ('relatedRevItemsLink','eventLink','invoiceLink'),
        }),
        (_('File Attachment (optional)'), {
            'fields': ('attachment',)
        }),
        (_('Approval/Payment Status'), {
            'fields': ('submissionUserLink','currentlyHeldBy','received','receivedDate','accrualDate')
        }),
    )

    def user_full_name(self,obj):
        if obj.submissionUser:
            return obj.submissionUser.first_name + ' ' + obj.submissionUser.last_name
        return ''

    def get_admin_change_link(self,app_label, model_name, obj_id, name):
        url = reverse('admin:%s_%s_change' % (app_label, model_name),
                      args=(obj_id,))
        return format_html('<a href="%s">%s</a>' % (
            url, str(name)
        ))

    def relatedRevItemsLink(self,obj):
        link = []
        for item in obj.relatedItems or []:
            link.append(self.get_admin_change_link('financial','revenueitem',item.id,item.__str__()))
            return ', '.join(link)
    relatedRevItemsLink.allow_tags = True
    relatedRevItemsLink.short_description = _('Revenue Items')

    def eventLink(self,obj):
        if obj.eventregistration:
            sr = obj.eventregistration
            return 'Registration: ' + self.get_admin_change_link('core','registration',sr.registration.id,sr.registration.__str__())
        elif obj.event:
            s = obj.event
            return self.get_admin_change_link('core','event',s.id,s.__str__())
    eventLink.allow_tags = True
    eventLink.short_description = _('Series/Event')

    def invoiceLink(self,obj):
        ''' If vouchers app is enabled and there is a voucher, this will link to it. '''
        if hasattr(obj,'invoiceItem') and obj.invoiceItem:
            return self.get_admin_change_link('core','invoice',obj.invoiceItem.invoice.id,_('Invoice'))
    invoiceLink.allow_tags = True
    invoiceLink.short_description = _('Invoice')

    def submissionUserLink(self,obj):
        link = []

        if obj.submissionUser:
            u = obj.submissionUser
            link.append(self.get_admin_change_link('auth','user',u.id,u.get_full_name()))
        if obj.submissionDate:
            link.append(str(obj.submissionDate))
        return ', '.join(link)
    submissionUserLink.allow_tags = True
    submissionUserLink.short_description = _('Submitted')

    def markReceived(self, request, queryset):
        rows_updated = queryset.update(received=True)
        if rows_updated == 1:
            message_bit = "1 revenue item was"
        else:
            message_bit = "%s revenue items were" % rows_updated
        self.message_user(request, "%s successfully marked as received." % message_bit)
    markReceived.short_description = _('Mark Revenue Items as received')

    def markNotReceived(self, request, queryset):
        rows_updated = queryset.update(received=False)
        if rows_updated == 1:
            message_bit = "1 revenue item was"
        else:
            message_bit = "%s revenue items were" % rows_updated
        self.message_user(request, "%s successfully marked as not received." % message_bit)
    markNotReceived.short_description = _('Mark Revenue Items as not received')

    def save_model(self,request,obj,form,change):
        obj.submissionUser = request.user
        obj.save()


class LocationRentalInfoInline(admin.StackedInline):
    model = LocationRentalInfo
    extra = 1
    fields = (('rentalRate','applyRateRule'),('dayStarts','weekStarts','monthStarts'),('advanceDays','advanceDaysReference','priorDays','priorDaysReference'),'disabled')
    classes = ('collapse',)

    def has_add_permission(self, request, obj=None):
        return False


class RoomRentalInfoInline(admin.StackedInline):
    model = RoomRentalInfo
    extra = 1
    fields = (('rentalRate','applyRateRule'),('dayStarts','weekStarts','monthStarts'),('advanceDays','advanceDaysReference','priorDays','priorDaysReference'),'disabled')
    classes = ('collapse',)

    def has_add_permission(self, request, obj=None):
        return False


class StaffMemberWageInfoInline(admin.StackedInline):
    model = StaffMemberWageInfo
    min_num = 0
    extra = 0
    fields = (('category','rentalRate','applyRateRule'),('dayStarts','weekStarts','monthStarts'),('advanceDays','advanceDaysReference','priorDays','priorDaysReference'),'disabled')
    classes = ('collapse',)


def updateStaffCompensationInfo(self, request, queryset):
    '''
    This action is added to the list for instructors to permit bulk
    updating of compensation information for staff members.
    '''
    selected = request.POST.getlist(admin.ACTION_CHECKBOX_NAME)
    ct = ContentType.objects.get_for_model(queryset.model)
    return HttpResponseRedirect(reverse('updateCompensationRules') + "?ct=%s&ids=%s" % (ct.pk, ",".join(selected)))


updateStaffCompensationInfo.short_description = _('Update compensation rules')


class RepeatedExpenseRuleChildAdmin(PolymorphicChildModelAdmin):
    """ Base admin class for all child models """
    base_model = RepeatedExpenseRule
    readonly_fields = ('lastRun',)

    # By using these `base_...` attributes instead of the regular ModelAdmin `form` and `fieldsets`,
    # the additional fields of the child models are automatically added to the admin form.
    base_fieldsets = (
        (None, {
            'fields': ('rentalRate','applyRateRule','disabled','lastRun')
        }),
        (_('Generation rules'),{
            'fields': ('dayStarts','weekStarts','monthStarts',('advanceDays','advanceDaysReference'),('priorDays','priorDaysReference')),
        }),
        (_('Start and End Date (optional)'), {
            'fields': ('startDate','endDate')
        }),
    )

    def get_fieldsets(self, request, obj=None):
        ''' Override polymorphic default to put the subclass-specific fields first '''
        # If subclass declares fieldsets, this is respected
        if (hasattr(self, 'declared_fieldset') and self.declared_fieldsets) \
           or not self.base_fieldsets:
            return super(PolymorphicChildModelAdmin, self).get_fieldsets(request, obj)

        other_fields = self.get_subclass_fields(request, obj)

        if other_fields:
            return (
                (self.extra_fieldset_title, {'fields': other_fields}),
            ) + self.base_fieldsets
        else:
            return self.base_fieldsets


@admin.register(LocationRentalInfo)
class LocationRentalInfoAdmin(RepeatedExpenseRuleChildAdmin):
    base_model = LocationRentalInfo
    extra_fieldset_title = None


@admin.register(RoomRentalInfo)
class RoomRentalInfoAdmin(RepeatedExpenseRuleChildAdmin):
    base_model = RoomRentalInfo
    extra_fieldset_title = None


@admin.register(StaffMemberWageInfo)
class StaffMemberWageInfoAdmin(RepeatedExpenseRuleChildAdmin):
    base_model = StaffMemberWageInfo
    extra_fieldset_title = None


class GenericRepeatedExpenseAdminForm(ModelForm):

    paymentMethod = autocomplete.Select2ListCreateChoiceField(
        choice_list=get_method_list,
        required=False,
        widget=autocomplete.ListSelect2(url='paymentMethod-list-autocomplete')
    )

    def __init__(self,*args,**kwargs):
        super(GenericRepeatedExpenseAdminForm,self).__init__(*args,**kwargs)

        updatedChoices = RepeatedExpenseRule.RateRuleChoices.choices
        index = [x[0] for x in updatedChoices].index('D')
        updatedChoices = updatedChoices[:index] + (('D',_('Per day')),) + updatedChoices[index + 1:]
        self.fields.get('applyRateRule').choices = updatedChoices


@admin.register(GenericRepeatedExpense)
class GenericRepeatedExpenseAdmin(RepeatedExpenseRuleChildAdmin):
    base_model = GenericRepeatedExpense
    base_form = GenericRepeatedExpenseAdminForm

    base_fieldsets = (
        (None, {
            'fields': ('name','rentalRate','applyRateRule','disabled','lastRun')
        }),
        (_('Generation rules'),{
            'fields': ('dayStarts','weekStarts','monthStarts',('advanceDays','advanceDaysReference'),('priorDays','priorDaysReference')),
        }),
        (_('Expense item parameters'),{
            'fields': ('category','payToUser','payToLocation','payToName',)
        }),
        (_('Start and End Date (optional)'), {
            'fields': ('startDate','endDate'),
            'classes': ('collapse',)
        }),
        (_('Automated approval'),{
            'fields': ('markApproved','markPaid','paymentMethod',),
            'classes': ('collapse',)
        }),
    )


@admin.register(RepeatedExpenseRule)
class RepeatedExpenseRuleAdmin(PolymorphicParentModelAdmin):
    base_model = RepeatedExpenseRule
    child_models = (LocationRentalInfo,RoomRentalInfo,StaffMemberWageInfo,GenericRepeatedExpense)
    polymorphic_list = True

    list_display = ('ruleName','applyRateRule','rentalRate','_enabled','lastRun')
    list_filter = (PolymorphicChildModelFilter,'applyRateRule','startDate','endDate','lastRun')

    actions = ('disableRules','enableRules','generateExpenses')

    def _enabled(self,obj):
        ''' Change the label '''
        return obj.disabled is False
    _enabled.short_description = _('Enabled')
    _enabled.boolean = True

    def disableRules(self, request, queryset):
        rows_updated = queryset.update(disabled=True)
        if rows_updated == 1:
            message_bit = "1 rule was"
        else:
            message_bit = "%s rules were" % rows_updated
        self.message_user(request, "%s successfully disabled." % message_bit)
    disableRules.short_description = _('Disable selected repeated expense rules')

    def enableRules(self, request, queryset):
        rows_updated = queryset.update(disabled=False)
        if rows_updated == 1:
            message_bit = "1 rule was"
        else:
            message_bit = "%s rules were" % rows_updated
        self.message_user(request, "%s successfully enabled." % message_bit)
    enableRules.short_description = _('Enable selected repeated expense rules')

    def generateExpenses(self, request, queryset):
        rule_count = len(queryset)
        generate_count = 0
        for q in queryset:
            generate_count += q.generateExpenses(request=request)
        message_bit = 'Successfully generated %s expense item%s from %s rule%s.' % (
            generate_count, 's' if generate_count != 1 else '', rule_count, 's' if rule_count != 1 else '')
        self.message_user(request, message_bit)

    generateExpenses.short_description = _('Generate expenses for selected repeated expense rules')


admin.site.register(ExpenseCategory)
admin.site.register(RevenueCategory)

# This adds inlines to Location and Room without subclassing
admin.site._registry[Location].inlines.insert(0,LocationRentalInfoInline)
admin.site._registry[Room].inlines.insert(0,RoomRentalInfoInline)
admin.site._registry[StaffMember].inlines.insert(0,StaffMemberWageInfoInline)
admin.site._registry[Instructor].inlines.insert(0,StaffMemberWageInfoInline)
admin.site._registry[StaffMember].actions.insert(0,updateStaffCompensationInfo)
admin.site._registry[Instructor].actions.insert(0,updateStaffCompensationInfo)
